@@ -1,4 +1,4 @@
-import re
+import regex as re
 import sys
 from sly import Lexer
 from sly.lex import Token
@@ -25,10 +25,13 @@ class JSON5Token(Token):
 
 
 class JSONLexer(Lexer):
+    regex_module = re
     reflags = re.DOTALL
     tokens = {LBRACE, RBRACE,
               LBRACKET, RBRACKET,
               DOUBLE_QUOTE_STRING, SINGLE_QUOTE_STRING,
+              UNTERMINATED_DOUBLE_QUOTE_STRING,
+              UNTERMINATED_SINGLE_QUOTE_STRING,
               NAME,
               COMMA,
               BLOCK_COMMENT,
@@ -78,7 +81,7 @@ class JSONLexer(Lexer):
     OCTAL = r'(0\d+|0o\d+)'
     FLOAT = r'(\d+\.\d*)|(\d*\.\d+)'      # 23.45
     INTEGER = r'\d+'
-    NAME = r'[\w_\$\\]([\w_\d\$\\])*'
+    NAME = r'[\w_\$\\]([\w_\d\$\\\p{Pc}\p{Mn}\p{Mc}\u200C\u200D])*'
 
     NAME['true'] = TRUE
     NAME['false'] = FALSE
@@ -86,6 +89,9 @@ class JSONLexer(Lexer):
     NAME['Infinity'] = INFINITY
     NAME['NaN'] = NAN
 
+    UNTERMINATED_DOUBLE_QUOTE_STRING = r'"(?:[^"\\]|\\.)*'
+    UNTERMINATED_SINGLE_QUOTE_STRING = r"'(?:[^'\\]|\\.)*"
+
     def error(self, t):
         raise JSON5DecodeError(f'Illegal character {t.value[0]!r} at index {self.index}', None)
 
